package com.vdv.stuntcar;

import android.content.Context;
import android.content.res.AssetFileDescriptor;
import android.content.res.Resources;
import android.media.AudioFormat;
import android.media.AudioManager;
import android.media.AudioTrack;

import com.vdv.sc.R;

import java.io.IOException;
import java.io.InputStream;
import java.net.ContentHandler;

class SoundManager extends Thread {

	private static final int AMIGA_PAL_HZ = 3546895;

	static final class Effect {
		int[] mSamples;
		int mEffect;// 0
		short mPeriod;// 8
		short mVolume_fp6;// 10
		byte mChannel;// 12

		Effect(final int id,final short period, final short volume, final byte channel) {
			mEffect = id;
			mPeriod = period;
			mVolume_fp6 = volume;
			mChannel = channel;
		}
	}

	/** This sound effect is only used when defining keys (press F1 while games paused) */
	static final int SOUND_CHIME = 0;
	static final int SOUND_WRECK = 1;
	static final int SOUND_HIT_CAR = 2;
	static final int SOUND_GROUNDED = 3;
	static final int SOUND_CREAK = 4;
	static final int SOUND_SMASH = 5;
	static final int SOUND_OFF_ROAD = 6;
	static final int SOUND_TICK_OVER = 7;

	// 12 bytes per effect
	static final Effect[] effects = {
			/*
			 * This sound effect is only used
			 * when defining keys (press F1 while games paused)
			 */
			// static final Effect chime =
			new Effect( 2436, (short) 150, (short) 30, (byte) 1 ),
			// static final Effect wreck =
			new Effect( 9032, (short) 180, (short) 64, (byte) 1 ),
			// static final Effect hit_car =
			new Effect( 8014, (short) 238, (short) 56, (byte) 1 ),
			// static final Effect grounded =
			new Effect( 3108, (short) 400, (short) 50, (byte) 1 ),
			// static final Effect creak =
			new Effect( 5170, (short) 238, (short) 64, (byte) 2 ),
			// static final Effect smash =
			new Effect( 8430, (short) 280, (short) 64, (byte) 3 ),
			// static final Effect off_road =
			new Effect( 7120, (short) 500, (short) 64, (byte) 1 ),
			// static final Effect tick_over =
			new Effect( 3172, (short) 300, (short) 48, (byte) 0 )
	};

	// java specific
	/** Sample rate of audio resources */
	private static final int SAMPLE_RATE = 11025;

	private static SoundManager mInstance = null;

	private static volatile boolean mIsWork = false;
	private static volatile boolean mIsPause = false;

	/** Buffer for engine sound. int type for speed */
	private static final int[][] mEngine = new int[ 8 ][];

	/** Sound buffer for pitched engine sound data */
	private static int[] mEngineSoundBuff = null;

	/** Engine sound sample count in bytes after pitching x2 */
	private static int mEngineSoundBytesCount_x2 = 0;

	/** Sound buffer for pitched sound effect data */
	private static int[] mPitchedEffectSoundBuff = null;

	private static int mPitchedEffectSampleCount = 0;

	/** Buffer to send sound into the DataLine */
	private static byte[] mMixBuff = null;

	/** Gain for the left channel, 64 is maximum, 0 is mute */
	private static int mGainLeft_fp6 = 0;

	/** Gain for the right channel, 256 is maximum, 0 is mute */
	private static int mGainRight_fp6 = 0;

	/** Current mix position */
	private static int mMixPosition = 0;

	/** Current sound effect id */
	private static int mCurrentSoundIndex = -1;
	// private static int mPreviousEngineSoundIndex = -1;
	private static volatile int mCurrentPeriod = AMIGA_PAL_HZ / SAMPLE_RATE;

	/**
	 *
	 * @param effectId effect index for {@link #effects}, as d0 byte
	 */
	static void soundEffect(final int effectId) {
		synchronized( mInstance ) {
			mIsPause = false;
			mCurrentSoundIndex = effectId;
			mInstance.notify();
		}
	}

	// sound_off
	static void soundOff() {
		// Game.jLogErrLn( "soundOff" );
		// int i = mEngineSoundBuff.length;
		// do{ mEngineSoundBuff[ --i ] = 0; } while( i > 0 );
		mIsPause = true;
	}

	// java
	/**
	 *
	 * @param period period. 124 <= period <= 0x3FFE. in d3.w
	 */
	static void pitchEngineSound(int period) {
		// Game.jLogErrLn( "pitchEngineSound: " + sd3 );
		mCurrentPeriod = period;
	}

	/** Private to make only one manager */
	private SoundManager() {
	}

	/** Load sound resource data and convert it to 16 bits sound */
	private static int[] getClip(final Resources r, final int nameId, final int factorFp6) throws IOException {
		try( final AssetFileDescriptor ais = r.openRawResourceFd( nameId );
			 final InputStream is = ais.createInputStream() )
		{
			final long skipped = is.skip( 44 );
			final int[] data = new int[ (int)ais.getLength() - 44 ];
			int count = 0;
			int sample;
			while( (sample = is.read()) >= 0 ) {// unsigned 8 to signed 16
				data[ count++ ] = (((sample - 128) << 7) * factorFp6) >> 6;
			}
			return data;
		} catch(final Exception e) {
			throw new IOException( e.getMessage() );
		}
	}

	/** Initiate the sound manager */
	static int init(final Resources r) {
		if( mInstance != null ) {
			return -1;
		}
		mInstance = new SoundManager();
		mInstance.setName("Sound");
		try {
			final int gain_fp6 = effects[ SOUND_TICK_OVER ].mVolume_fp6;
			effects[ SOUND_TICK_OVER ].mSamples = mEngine[ 0 ] = getClip( r, R.raw.tickover, gain_fp6 );
			mEngine[ 1 ] = getClip( r, R.raw.enginepitch2, gain_fp6 );
			mEngine[ 2 ] = getClip( r, R.raw.enginepitch3, gain_fp6 );
			mEngine[ 3 ] = getClip( r, R.raw.enginepitch4, gain_fp6 );
			mEngine[ 4 ] = getClip( r, R.raw.enginepitch5, gain_fp6 );
			mEngine[ 5 ] = getClip( r, R.raw.enginepitch6, gain_fp6 );
			mEngine[ 6 ] = getClip( r, R.raw.enginepitch7, gain_fp6 );
			mEngine[ 7 ] = getClip( r, R.raw.enginepitch8, gain_fp6 );
			effects[ SOUND_CHIME ].mSamples = getClip( r, R.raw.chime, 64 );
			effects[ SOUND_WRECK ].mSamples = getClip( r, R.raw.wreck, 64 );
			effects[ SOUND_HIT_CAR ].mSamples = getClip( r, R.raw.hitcar, 64 );
			effects[ SOUND_GROUNDED ].mSamples = getClip( r, R.raw.grounded, 64 );
			effects[ SOUND_CREAK ].mSamples = getClip( r, R.raw.creak, 64 );
			effects[ SOUND_SMASH ].mSamples = getClip( r, R.raw.smash, 64 );
			effects[ SOUND_OFF_ROAD ].mSamples = getClip( r, R.raw.offroad, 64 );
			// uses pitch only >= 1
			mEngineSoundBuff = new int[ effects[ SOUND_TICK_OVER ].mSamples.length ];
			mEngineSoundBytesCount_x2 = mEngineSoundBuff.length << 2;// 2 bytes per sample * 2channels
			int samples = 0;
			for( final Effect effect : effects ) {
				samples = Math.max( samples, effect.mSamples.length );
			}
			mPitchedEffectSoundBuff = new int[ samples ];
			mMixBuff = new byte[ mEngineSoundBytesCount_x2 ];
			mInstance.start();
		} catch(final Exception e) {
			e.printStackTrace();
			return -1;
		}
		return 0;
	}

	static void pause() {
        mIsPause = true;
	}

	static void restart() {
        synchronized( mInstance ) {
            mIsPause = false;
            mInstance.notify();
        }
	}

	// --- to pitch engine sound
	private static final int ACCURACY = 16;
	private static final int MASK = (1 << ACCURACY) - 1;
	private static int mTf = 0;
	private static int mT = 0;
	private static int mInT = 0;
	private static int mPrevious = 0;

	/**
	 * Pitch sound.
	 *
	 * @param inSnd - input sound buffer
	 * @param outSnd - buffer with pitched sound
	 * @param inSampleRate - numerator to calculate the pitch ratio
	 * @param outSampleRate - denominator to calculate the pitch ratio
	 * @return samples in the output buffer
	 */
	private static int pitch(final int[] inSnd, final int[] outSnd, int inSampleRate, int outSampleRate) {
		final int quotient = (int)(((long)inSampleRate << ACCURACY) / (long)outSampleRate);
		int dT_fixed = (quotient & MASK);
		int dT = (quotient >> ACCURACY);
		int out_count = 0;
		int s = 0;
		do {
			while( mInT <= mT ) {
				mPrevious = inSnd[ s++ ];
				mInT++;
				if( s >= inSnd.length ) {
					if( mInT < mT ) {
						mT = mT - mInT;
						mInT = 0;
					} else {
						mInT = mInT - mT;
						mT = 0;
					}
					return out_count;
				}
			}
			int t = (mPrevious + ((mTf * (inSnd[ s ] - mPrevious)) >> ACCURACY));
			outSnd[ out_count++ ] = t;
			t = mTf + dT_fixed;
			mT = mT + dT + (t >> ACCURACY);
			mTf = (t & MASK);
		}
		while( out_count < outSnd.length );
		if( mInT < mT ) {
			mT = mT - mInT;
			mInT = 0;
		} else {
			mInT = mInT - mT;
			mT = 0;
		}
		return out_count;
	}
	// --- end pitch engine sound

	/** Sound thread */
	public void run() {
		AudioTrack line = null;
		try {
			final int min_buf_size = AudioTrack.getMinBufferSize( SAMPLE_RATE,
					AudioFormat.CHANNEL_OUT_STEREO, AudioFormat.ENCODING_PCM_16BIT );
			line = new AudioTrack( AudioManager.STREAM_MUSIC, SAMPLE_RATE,
					AudioFormat.CHANNEL_OUT_STEREO, AudioFormat.ENCODING_PCM_16BIT,
					min_buf_size, AudioTrack.MODE_STREAM );
			line.play();

			mIsWork = true;
			mIsPause = true;

			while( mIsWork ) {

				while( mIsPause ) {
					synchronized( mInstance ) {
						try { wait(); } catch( final Exception ignored ) {}
					}
				}
				// pitch
				// Calculate sound index that will give period < 256
				int sound_index = 6;
				int period = mCurrentPeriod;
				while( period >= 256 ) {
					period >>>= 1;
					if( --sound_index < 0 ) {
						sound_index = 0;
					}
				}
				int frequency = AMIGA_PAL_HZ / period;
				/* if( mPreviousEngineSoundIndex != engine_sound_index ) {
					// Attempt to start at the same position through as a previous sound
					if( engine_sound_index > mPreviousEngineSoundIndex ) {
						mCurrentSample >>>= 1;
					} else {
						mCurrentSample <<= 1;
					}
					mPreviousEngineSoundIndex = engine_sound_index;// are pitches greater than 5 required?
				} */
				mEngineSoundBytesCount_x2 = mEngine[ sound_index ].length << 2;
				// System.arraycopy( mEngine[ sound_index ], 0, mEngineSoundBuff, 0, mEngine[ engine_sound_index ].length );
				mEngineSoundBytesCount_x2 = pitch( mEngine[ sound_index ], mEngineSoundBuff, frequency, SAMPLE_RATE ) << 2;

				// new effect
				sound_index = mCurrentSoundIndex;
				mCurrentSoundIndex = -1;
				if( sound_index >= 0 ) {
					mMixPosition = 0;
					// mIsLoop = false;
					final Effect effect = effects[ sound_index ];
					mGainRight_fp6 = mGainLeft_fp6 = effect.mVolume_fp6;
					frequency = AMIGA_PAL_HZ / effect.mPeriod;
					mPitchedEffectSampleCount = pitch( effect.mSamples, mPitchedEffectSoundBuff, frequency, SAMPLE_RATE );
				}

				int i = 0;
				do {// mix 2 mono 16 bits channels to one stereo 16 bits channel
					if( ++mMixPosition >= mPitchedEffectSampleCount ) {
						mMixPosition = 0;
						// if( ! mIsLoop ) {// mute
						mGainRight_fp6 = mGainLeft_fp6 = 0;
						// }
					}
					final int sample = mPitchedEffectSoundBuff[ mMixPosition ];
					final int engine_sample = mEngineSoundBuff[ i >> 2 ];
					// left channel
					int mix = engine_sample + ((mGainLeft_fp6 * sample) >> 6);
					mMixBuff[ i++ ] = (byte) mix;
					mMixBuff[ i++ ] = (byte) (mix >> 8);
					// right channel
					mix = engine_sample + ((mGainRight_fp6 * sample) >> 6);
					mMixBuff[ i++ ] = (byte) mix;
					mMixBuff[ i++ ] = (byte) (mix >> 8);
				} while( i < mEngineSoundBytesCount_x2 );
				line.write( mMixBuff, 0, mEngineSoundBytesCount_x2 );
				//mLine.drain();
			}
		} catch(final Exception e) {
			e.printStackTrace();
		} finally {
			if( line != null ) {
				line.flush();
				line = null;
			}
		}
	}
}